signature GAMESTATE =
sig

datatype cell = 
	 RED
       | GREEN
       | BLUE
       | GRAY

datatype dir = N | E | S | W

val w : int
val h : int

type effect
val noeffect : effect

type state
val initial : cell list -> state
val update : state * dir -> (state * effect) option 

type surface
val drawTo : state * effect * surface * surface -> unit
val getScore : state -> int

val drawcursor : bool ref
end

(* ... *)

functor GameState(structure Sdl : SDL) :>
	GAMESTATE 
	where type surface = Sdl.surface =
struct
type surface = Sdl.surface

datatype cell = 
         RED
       | GREEN
       | BLUE
       | GRAY

type state = {x:int, y:int, cells:cell option list, q: cell list, score : int}
val w = 12
val h = 10

type effect = (int * int) list
val noeffect = []

datatype dir = N | E | S | W
fun dir2v N = (0,~1)
  | dir2v E = (1,0)
  | dir2v S = (0,1)
  | dir2v W = (~1,0)
fun vplus ((x,y),(u,v)) = (x+u,y+v)
fun inbounds (x,y) = x >= 0 andalso x < w andalso y >= 0 andalso y < h


fun sub cells (x, y) = List.nth (cells, x + y * w)
fun listmodify 0 f (h::tl) = f h :: tl
  | listmodify n f (h::tl) = h :: listmodify (n-1) f tl
  | listmodify _ _ _ = raise Match

fun modify (x, y) f cells = listmodify (x + y * w) f cells
	 
fun initial (hd::q) = {x = 0, y = h-1, cells = modify (0,h-1) (fn _ => SOME hd)
						      (List.tabulate (w * h, fn _ => NONE)), q = q, score=0}
  | initial _ = raise Match

fun listelm x [] = false
  | listelm x (y::tl) = if x = y then true else listelm x tl

fun getClique p cells =
    let
	val target = sub cells p
	(* dfs trylist ignorelist = goodlist *)
	fun children h = List.filter inbounds (map (fn d => vplus(h, dir2v d)) [N, E, S, W])
	fun dfs [] succ = succ
	  | dfs (h::tl) succ = if sub cells h = target andalso not (listelm h succ)
			       then dfs (children h @ tl) (h :: succ)
			       else dfs tl succ
	val answer = dfs [p] []
	fun pos2str (x, y) = "(" ^ Int.toString x ^ ", " ^ Int.toString y ^ ")"
(*	val _ = print ((StringUtil.delimit ", " (map pos2str answer)) ^ "\n") *)
    in
	answer
    end

exception NoFree
fun update ({q = [], ...}, _) = NONE
  | update ({x, y, cells, q = h::tl, score}, dir) = 
    let
	val dirv = dir2v dir
	fun getnewp (x,y) dirv = 
	    if inbounds(x,y)
	    then 
		case sub cells (x,y) of NONE => (x,y)
				      | SOME _ => getnewp (vplus (dirv, (x,y))) dirv
	    else raise NoFree
	val (newp as (newx, newy)) = getnewp (vplus (dirv, (x,y))) dirv
	val newcells =  modify newp (fn _ => SOME h) cells
	val clique = getClique newp newcells
	val newscore = score + 10 * (length clique) - 9
    in
	SOME ({x = newx, y = newy, cells = newcells, q = tl, score = newscore}, 
	      clique)
    end
	handle NoFree => NONE

val drawcursor = ref true

fun drawTo ({x, y, cells, q, score}, clique, sprites, surface) =
    let
	fun cell2int RED = 0
	  | cell2int GREEN = 1
	  | cell2int GRAY = 2
	  | cell2int BLUE = 3

	fun blit (a,b) (x,y) = Sdl.blit(sprites, surface, 32 * a, 32 * b, 32, 32, x, y)
	fun normal_blit (a) (x,y) = blit (a, 0) (32 + x * 32, 32 + y * 32)
	fun bright_blit (a) (x,y) = blit (a, 1) (32 + x * 32, 32 + y * 32)
	fun qblit a x = blit (cell2int a, 0) (32 + x * 32, 32 + (h + 1) * 32)

	fun draw_cell f NONE p = ()
	  | draw_cell f (SOME c) p = f (cell2int c) p 

	fun next (x,y) = if x + 1 = w then (0, y+1) else (x+1,y)
	fun go [] p = ()
	  | go (cell::tl) (p as (px, py)) = 
	    let

	    in
		draw_cell (if length clique > 1 andalso listelm p clique
			   then bright_blit
			   else normal_blit) cell p;
		go tl (next p)
	    end
	fun goq [] n = ()
	  | goq (a::tl) n = if n = 10 then () else (qblit a n; goq tl (n+1))
    in
	(* draw some background rects *)
	Sdl.fillrect (surface, 32, 32, w * 32, h * 32, 0wxafafaf);
	Sdl.fillrect (surface, 32, 32 + (h+1) * 32, w * 32, 32, 0wxafafaf);
	(* draw main board *)
	go cells (0,0);
	if (!drawcursor) then blit (0,2) (32 + x * 32, 32 + y * 32) else ();
	(* draw tile queue *)
	goq q 0
    end

fun getScore ({score,...} : state) = score

end
